#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "GBACompress.h"

#define GBA_E_OK					0
#define GBA_E_ERROR					(-1)
#define GBA_E_INVALID_ARGUMENT		(-2)
#define GBA_E_OUT_OF_MEMORY			(-3)

#define SWD_HSIZE   65536ul

#define GBA_UNUSED(parm)  (parm = parm)
#define GBA_MIN(a,b)		((a) <= (b) ? (a) : (b))

#define UINT32_C(c)	 c ## U
#define BYTE_C(x)	   ((unsigned char) (x))

/***********************************************************************
//
************************************************************************/

#define N	   (1024*1024ul)	   /* size of ring buffer */
#define SWD_USE_MALLOC
#define THRESHOLD	   1		   /* lower limit for match length */
#define F			2048		   /* upper limit for match length */

#define M2_MAX_OFFSET				 0xd00

/***********************************************************************
//
************************************************************************/

struct gba_compress_config_t
{
	int bb_endian;
	int bb_size;
	u32 max_offset;
	u32 max_match;
	int s_level;
	int h_level;
	int p_level;
	int c_flags;
	u32 m_size;
};

typedef struct
{
	void *callback;
	void * user;
}
gba_progress_callback_t;

#define gba_progress_callback_p gba_progress_callback_t *

typedef struct
{
	int init;

	u32 look;		  /* bytes in lookahead buffer */

	u32 m_len;
	u32 m_off;

	u32 last_m_len;
	u32 last_m_off;

	const u8 *bp;
	const u8 *ip;
	const u8 *in;
	const u8 *in_end;
	u8 *out;

	u32 bb_b;
	unsigned bb_k;
	unsigned bb_c_endian;
	unsigned bb_c_s;
	unsigned bb_c_s8;
	u8 *bb_p;
	u8 *bb_op;

	struct gba_compress_config_t conf;
	u32 * result;

	gba_progress_callback_p cb;

	u32 textsize;	  /* text size counter */
	u32 codesize;	  /* code size counter */
	u32 printcount;	/* counter for reporting progress every 1K bytes */

	/* some stats */
	unsigned long lit_bytes;
	unsigned long match_bytes;
	unsigned long rep_bytes;
	unsigned long lazy;
}
GBA_COMPRESS_T;

#define getbyte(c)  ((c).ip < (c).in_end ? *((c).ip)++ : (-1))

/***********************************************************************
//
************************************************************************/

#  define SWD_N			 N
#  define SWD_F			 F
#  define SWD_THRESHOLD	 THRESHOLD

/* unsigned type for dictionary access - don't waste memory here */
typedef u32		 swd_uint;
#define UINT_MIN	  UINT_MAX
#define SWD_UINT(x)		 ((swd_uint)(x))


#define SWD_MAX_CHAIN	 2048

#  define HEAD3(b,p) \
	(((0x9f5f*(((((u32)b[p]<<5)^b[p+1])<<5)^b[p+2]))>>5) & (SWD_HSIZE-1))

#	define HEAD2(b,p)	  (b[p] ^ ((unsigned)b[p+1]<<8))
#  define NIL2			  UINT_MIN

typedef struct
{
/* public - "built-in" */
	u32 n;
	u32 f;
	u32 threshold;

/* public - configuration */
	u32 max_chain;
	u32 nice_length;
	int use_best_off;
	u32 lazy_insert;

/* public - output */
	u32 m_len;
	u32 m_off;
	u32 look;
	int b_char;

/* semi public */
	GBA_COMPRESS_T *c;
	u32 m_pos;

/* private */
	const u8 *dict;
	const u8 *dict_end;
	u32 dict_len;

/* private */
	u32 ip;				/* input pointer (lookahead) */
	u32 bp;				/* buffer pointer */
	u32 rp;				/* remove pointer */
	u32 b_size;

	unsigned char *b_wrap;

	u32 node_count;
	u32 first_rp;

	unsigned char *b;
	swd_uint *head3;
	swd_uint *succ3;
	swd_uint *best3;
	swd_uint *llen3;
	swd_uint *head2;
}
gba_swd_t;


/* Access macro for head3.
 * head3[key] may be uninitialized if the list is emtpy,
 * but then its value will never be used.
 */
#define s_head3(s,key)		s->head3[key]


/***********************************************************************
//
************************************************************************/

void *gba_alloc(u32 nelems, u32 size)
{
	return calloc(nelems,size);
}

void gba_free(void *p)
{
	free(p);
}

void *gba_memset(void *s, int c, u32 len)
{
	return memset(s,c,len);
}

void *gba_memcpy(void *dest, const void *src, u32 len)
{
	u8 *p1 = (u8 *) dest;
	const u8 *p2 = (const u8 *) src;

	if (len <= 0 || p1 == p2)
		return dest;
	do
		*p1++ = *p2++;
	while (--len > 0);
	return dest;
}

static void swd_initdict(gba_swd_t *s, const u8 *dict, u32 dict_len)
{
	s->dict = s->dict_end = NULL;
	s->dict_len = 0;

	if (!dict || dict_len <= 0)
		return;

	if (dict_len > s->n)
	{
		dict += dict_len - s->n;
		dict_len = s->n;
	}

	s->dict = dict;
	s->dict_len = dict_len;
	s->dict_end = dict + dict_len;
	gba_memcpy(s->b,dict,dict_len);
	s->ip = dict_len;
}


static void swd_insertdict(gba_swd_t *s, u32 node, u32 len)
{
	u32 key;

	s->node_count = s->n - len;
	s->first_rp = node;

	while (len-- > 0)
	{
		key = HEAD3(s->b,node);
		s->succ3[node] = s_head3(s,key);
		s->head3[key] = SWD_UINT(node);
		s->best3[node] = SWD_UINT(s->f + 1);
		s->llen3[key]++;
//		assert(s->llen3[key] <= s->n);

		key = HEAD2(s->b,node);
		s->head2[key] = SWD_UINT(node);

		node++;
	}
}


/***********************************************************************
//
************************************************************************/

static int swd_init(gba_swd_t *s, const u8 *dict, u32 dict_len)
{
	u32 i = 0;
	int c = 0;

	if (s->n == 0)
		s->n = SWD_N;
	if (s->f == 0)
		s->f = SWD_F;
	s->threshold = SWD_THRESHOLD;
	if (s->n > SWD_N || s->f > SWD_F)
		return GBA_E_INVALID_ARGUMENT;

	s->b = (unsigned char *) gba_alloc(s->n + s->f + s->f, 1);
	s->head3 = (swd_uint *) gba_alloc(SWD_HSIZE, sizeof(*s->head3));
	s->succ3 = (swd_uint *) gba_alloc(s->n + s->f, sizeof(*s->succ3));
	s->best3 = (swd_uint *) gba_alloc(s->n + s->f, sizeof(*s->best3));
	s->llen3 = (swd_uint *) gba_alloc(SWD_HSIZE, sizeof(*s->llen3));
	if (!s->b || !s->head3  || !s->succ3 || !s->best3 || !s->llen3)
		return GBA_E_OUT_OF_MEMORY;

	s->head2 = (swd_uint *) gba_alloc(UINT32_C(65536), sizeof(*s->head2));
	if (!s->head2)
		return GBA_E_OUT_OF_MEMORY;

	/* defaults */
	s->max_chain = SWD_MAX_CHAIN;
	s->nice_length = s->f;
	s->use_best_off = 0;
	s->lazy_insert = 0;

	s->b_size = s->n + s->f;
	if (s->b_size + s->f >= UINT_MIN)
		return GBA_E_ERROR;
	s->b_wrap = s->b + s->b_size;
	s->node_count = s->n;

	gba_memset(s->llen3, 0, sizeof(s->llen3[0]) * SWD_HSIZE);

	gba_memset(s->head2, 0xff, sizeof(s->head2[0]) * UINT32_C(65536));
//	assert(s->head2[0] == NIL2);

	s->ip = 0;
	swd_initdict(s,dict,dict_len);
	s->bp = s->ip;
	s->first_rp = s->ip;

//	assert(s->ip + s->f <= s->b_size);
	s->look = (u32) (s->c->in_end - s->c->ip);
	if (s->look > 0)
	{
		if (s->look > s->f)
			s->look = s->f;
		gba_memcpy(&s->b[s->ip],s->c->ip,s->look);
		s->c->ip += s->look;
		s->ip += s->look;
	}

	if (s->ip == s->b_size)
		s->ip = 0;

	if (s->look >= 2 && s->dict_len > 0)
		swd_insertdict(s,0,s->dict_len);

	s->rp = s->first_rp;
	if (s->rp >= s->node_count)
		s->rp -= s->node_count;
	else
		s->rp += s->b_size - s->node_count;

	GBA_UNUSED(i);
	GBA_UNUSED(c);
	return GBA_E_OK;
}

static
void swd_exit(gba_swd_t *s)
{
	/* free in reverse order of allocations */
	gba_free(s->head2); s->head2 = NULL;
	gba_free(s->llen3); s->llen3 = NULL;
	gba_free(s->best3); s->best3 = NULL;
	gba_free(s->succ3); s->succ3 = NULL;
	gba_free(s->head3); s->head3 = NULL;
	gba_free(s->b); s->b = NULL;
}


#define swd_pos2off(s,pos) \
	(s->bp > (pos) ? s->bp - (pos) : s->b_size - ((pos) - s->bp))


/***********************************************************************
//
************************************************************************/

static 
void swd_getbyte(gba_swd_t *s)
{
	int c;

	if ((c = getbyte(*(s->c))) < 0)
	{
		if (s->look > 0)
			--s->look;
	}
	else
	{
		s->b[s->ip] = BYTE_C(c);
		if (s->ip < s->f)
			s->b_wrap[s->ip] = BYTE_C(c);
	}
	if (++s->ip == s->b_size)
		s->ip = 0;
	if (++s->bp == s->b_size)
		s->bp = 0;
	if (++s->rp == s->b_size)
		s->rp = 0;
}


/***********************************************************************
// remove node from lists
************************************************************************/

static 
void swd_remove_node(gba_swd_t *s, u32 node)
{
	if (s->node_count == 0)
	{
		u32 key;

		key = HEAD3(s->b,node);
//		assert(s->llen3[key] > 0);
		--s->llen3[key];

		key = HEAD2(s->b,node);
//		assert(s->head2[key] != NIL2);
		if ((u32) s->head2[key] == node)
			s->head2[key] = NIL2;
	}
	else
		--s->node_count;
}


/***********************************************************************
//
************************************************************************/

static
void swd_accept(gba_swd_t *s, u32 n)
{
//	assert(n <= s->look);

	if (n > 0) do
	{
		u32 key;

		swd_remove_node(s,s->rp);

		/* add bp into HEAD3 */
		key = HEAD3(s->b,s->bp);
		s->succ3[s->bp] = s_head3(s,key);
		s->head3[key] = SWD_UINT(s->bp);
		s->best3[s->bp] = SWD_UINT(s->f + 1);
		s->llen3[key]++;
//		assert(s->llen3[key] <= s->n);

		/* add bp into HEAD2 */
		key = HEAD2(s->b,s->bp);
		s->head2[key] = SWD_UINT(s->bp);

		swd_getbyte(s);
	} while (--n > 0);
}


/***********************************************************************
//
************************************************************************/

static
void swd_search(gba_swd_t *s, u32 node, u32 cnt)
{
	const unsigned char *p1;
	const unsigned char *p2;
	const unsigned char *px;
	u32 m_len = s->m_len;
	const unsigned char * b  = s->b;
	const unsigned char * bp = s->b + s->bp;
	const unsigned char * bx = s->b + s->bp + s->look;
	unsigned char scan_end1;

//	assert(s->m_len > 0);

	scan_end1 = bp[m_len - 1];
	for ( ; cnt-- > 0; node = s->succ3[node])
	{
		p1 = bp;
		p2 = b + node;
		px = bx;

//		assert(m_len < s->look);

		if (
			p2[m_len - 1] == scan_end1 &&
			p2[m_len] == p1[m_len] &&
			p2[0] == p1[0] &&
			p2[1] == p1[1])
		{
			u32 i;
//			assert(gba_memcmp(bp,&b[node],3) == 0);

			p1 += 2; p2 += 2;
			do {} while (++p1 < px && *p1 == *++p2);

			i = p1 - bp;

//			assert(gba_memcmp(bp,&b[node],i) == 0);

			if (i > m_len)
			{
				s->m_len = m_len = i;
				s->m_pos = node;
				if (m_len == s->look)
					return;
				if (m_len >= s->nice_length)
					return;
				if (m_len > (u32) s->best3[node])
					return;
				scan_end1 = bp[m_len - 1];
			}
		}
	}
}


/***********************************************************************
//
************************************************************************/

static
int swd_search2(gba_swd_t *s)
{
	u32 key;

//	assert(s->look >= 2);
//	assert(s->m_len > 0);

	key = s->head2[ HEAD2(s->b,s->bp) ];
	if (key == NIL2)
		return 0;

//	assert(gba_memcmp(&s->b[s->bp],&s->b[key],2) == 0);

	if (s->m_len < 2)
	{
		s->m_len = 2;
		s->m_pos = key;
	}
	return 1;
}


/***********************************************************************
//
************************************************************************/

static
void swd_findbest(gba_swd_t *s)
{
	u32 key;
	u32 cnt, node;
	u32 len;

//	assert(s->m_len > 0);

	/* get current head, add bp into HEAD3 */
	key = HEAD3(s->b,s->bp);
	node = s->succ3[s->bp] = s_head3(s,key);
	cnt = s->llen3[key]++;
//	assert(s->llen3[key] <= s->n + s->f);
	if (cnt > s->max_chain && s->max_chain > 0)
		cnt = s->max_chain;
	s->head3[key] = SWD_UINT(s->bp);

	s->b_char = s->b[s->bp];
	len = s->m_len;
	if (s->m_len >= s->look)
	{
		if (s->look == 0)
			s->b_char = -1;

		s->m_off = 0;
		s->best3[s->bp] = SWD_UINT(s->f + 1);
	}
	else
	{
		if (swd_search2(s))
			if (s->look >= 3)
				swd_search(s,node,cnt);

		if (s->m_len > len)
			s->m_off = swd_pos2off(s,s->m_pos);

		s->best3[s->bp] = SWD_UINT(s->m_len);
	}

	swd_remove_node(s,s->rp);

	/* add bp into HEAD2 */
	key = HEAD2(s->b,s->bp);
	s->head2[key] = SWD_UINT(s->bp);
}

/***********************************************************************
//
************************************************************************/

static int init_match ( GBA_COMPRESS_T *c, gba_swd_t *s,const u8 *dict, u32 dict_len,u32 flags )
{
	int r;

//	assert(!c->init);

	c->init = 1;

	s->c = c;

	c->last_m_len = c->last_m_off = 0;

	c->textsize = c->codesize = c->printcount = 0;
	c->lit_bytes = c->match_bytes = c->rep_bytes = 0;
	c->lazy = 0;

	r = swd_init(s,dict,dict_len);
	if (r != GBA_E_OK)
	{
		swd_exit(s);
		return r;
	}

	s->use_best_off = (flags & 1) ? 1 : 0;
	return GBA_E_OK;
}


/***********************************************************************
//
************************************************************************/

static int find_match ( GBA_COMPRESS_T *c, gba_swd_t *s, u32 this_len, u32 skip )
{
//	assert(c->init);

	if (skip > 0)
	{
//		assert(this_len >= skip);
		swd_accept(s, this_len - skip);
		c->textsize += this_len - skip + 1;
	}
	else
	{
//		assert(this_len <= 1);
		c->textsize += this_len - skip;
	}

	s->m_len = THRESHOLD;
	swd_findbest(s);
	c->m_len = s->m_len;
	c->m_off = s->m_off;

	swd_getbyte(s);

	if (s->b_char < 0)
	{
		c->look = 0;
		c->m_len = 0;
		swd_exit(s);
	}
	else
	{
		c->look = s->look + 1;
	}
	c->bp = c->ip - c->look;

	if (c->cb && c->textsize > c->printcount)
	{
//		(*c->cb->callback)(c->textsize,c->codesize,3,c->cb->user);
		c->printcount += 1024;
	}

	return GBA_E_OK;
}


/***********************************************************************
// bit buffer
************************************************************************/

static int bbConfig(GBA_COMPRESS_T *c, int endian, int bitsize)
{
	if (endian != -1)
	{
		if (endian != 0)
			return GBA_E_ERROR;
		c->bb_c_endian = endian;
	}
	if (bitsize != -1)
	{
		if (bitsize != 8 && bitsize != 16 && bitsize != 32)
			return GBA_E_ERROR;
		c->bb_c_s = bitsize;
		c->bb_c_s8 = bitsize / 8;
	}
	c->bb_b = 0; c->bb_k = 0;
	c->bb_p = NULL;
	c->bb_op = NULL;
	return GBA_E_OK;
}


static void bbWriteBits(GBA_COMPRESS_T *c)
{
	u8 *p = c->bb_p;
	u32 b = c->bb_b;

	p[0] = BYTE_C(b >>  0);
	if (c->bb_c_s >= 16)
	{
		p[1] = BYTE_C(b >>  8);
		if (c->bb_c_s == 32)
		{
			p[2] = BYTE_C(b >> 16);
			p[3] = BYTE_C(b >> 24);
		}
	}
}


static void bbPutBit(GBA_COMPRESS_T *c, unsigned bit)
{
	if (c->bb_k < c->bb_c_s)
	{
		if (c->bb_k == 0)
		{
			c->bb_p = c->bb_op;
			c->bb_op += c->bb_c_s8;
		}

		c->bb_b = (c->bb_b << 1) + bit;
		c->bb_k++;
	}
	else
	{
		bbWriteBits(c);
		c->bb_p = c->bb_op;
		c->bb_op += c->bb_c_s8;
		c->bb_b = bit;
		c->bb_k = 1;
	}
}


static void bbPutByte(GBA_COMPRESS_T *c, unsigned b)
{
	*c->bb_op++ = BYTE_C(b);
}


static void bbFlushBits(GBA_COMPRESS_T *c, unsigned filler_bit)
{
	if (c->bb_k > 0)
	{
		while (c->bb_k != c->bb_c_s)
			bbPutBit(c, filler_bit);
		bbWriteBits(c);
		c->bb_k = 0;
	}
	c->bb_p = NULL;
}



/***********************************************************************
//
************************************************************************/

static void code_prefix_ss11(GBA_COMPRESS_T *c, u32 i)
{
	if (i >= 2)
	{
		u32 t = 4;
		i += 2;
		do {
			t <<= 1;
		} while (i >= t);
		t >>= 1;
		do {
			t >>= 1;
			bbPutBit(c, (i & t) ? 1 : 0);
			bbPutBit(c, 0);
		} while (t > 2);
	}
	bbPutBit(c, (unsigned)i & 1);
	bbPutBit(c, 1);
}


static void code_match(GBA_COMPRESS_T *c, u32 m_len, const u32 m_off)
{
	unsigned m_low = 0;

	while (m_len > c->conf.max_match)
	{
		code_match(c, c->conf.max_match - 3, m_off);
		m_len -= c->conf.max_match - 3;
	}

	c->match_bytes += m_len;
	if (m_len > c->result[3])
		c->result[3] = m_len;
	if (m_off > c->result[1])
		c->result[1] = m_off;

	bbPutBit(c, 0);

	if (m_off == c->last_m_off)
	{
		bbPutBit(c, 0);
		bbPutBit(c, 1);
	}
	else
	{
		code_prefix_ss11(c, 1 + ((m_off - 1) >> 8));
		bbPutByte(c, (unsigned)m_off - 1);
	}
	m_len = m_len - 1 - (m_off > M2_MAX_OFFSET);
	if (m_len >= 4)
	{
		bbPutBit(c,0);
		bbPutBit(c,0);
		code_prefix_ss11(c, m_len - 4);
	}
	else
	{
		bbPutBit(c, m_len > 1);
		bbPutBit(c, (unsigned)m_len & 1);
	}

	c->last_m_off = m_off;
	GBA_UNUSED(m_low);
}

static void code_run(GBA_COMPRESS_T *c, const u8 *ii, u32 lit)
{
	if (lit == 0)
		return;
	c->lit_bytes += lit;
	if (lit > c->result[5])
		c->result[5] = lit;
	do
	{
		bbPutBit(c, 1);
		bbPutByte(c, *ii++);
	} while (--lit > 0);
}


/***********************************************************************
//
************************************************************************/

static int len_of_coded_match(GBA_COMPRESS_T *c, u32 m_len, u32 m_off)
{
	int b;
	if (m_len < 2 || (m_len == 2 && (m_off > M2_MAX_OFFSET)) || m_off > c->conf.max_offset)
		return -1;

	m_len = m_len - 2 - (m_off > M2_MAX_OFFSET);

	if (m_off == c->last_m_off)
	{
		b = 1 + 2;
	}
	else
	{
		b = 1 + 10;
		m_off = (m_off - 1) >> 8;
		while (m_off > 0)
		{
			b += 2;
			m_off >>= 1;
		}
	}

	b += 2;
	if (m_len < 3)
		return b;
	m_len -= 3;

	do {
		b += 2;
		m_len >>= 1;
	} while (m_len > 0);

	return b;
}


/***********************************************************************
//
************************************************************************/

#define assert_match(a,b,c)   ((void)0)

/***********************************************************************
//
************************************************************************/

int GBACompress( const u8 *in, u32 in_len,u8 *out, u32 *out_len,int level)
{
	void *cb = NULL;
	struct GBACompressConfig *conf = NULL;
	u32 *result = NULL;
	const u8 *ii;
	u32 lit;
	u32 m_len, m_off;
	GBA_COMPRESS_T c_buffer;
	GBA_COMPRESS_T * const c = &c_buffer;
#undef swd
	gba_swd_t the_swd;
#define swd (&the_swd)
	u32 result_buffer[16];
	int r;

	struct swd_config_t
	{
		unsigned try_lazy;
		u32 good_length;
		u32 max_lazy;
		u32 nice_length;
		u32 max_chain;
		u32 flags;
		u32 max_offset;
	};
	const struct swd_config_t *sc;
	static const struct swd_config_t swd_config[10] = {
		/* faster compression */
		{   0,   0,   0,   8,	4,   0,  48*1024L },
		{   0,   0,   0,  16,	8,   0,  48*1024L },
		{   0,   0,   0,  32,   16,   0,  48*1024L },
		{   1,   4,   4,  16,   16,   0,  48*1024L },
		{   1,   8,  16,  32,   32,   0,  48*1024L },
		{   1,   8,  16, 128,  128,   0,  48*1024L },
		{   2,   8,  32, 128,  256,   0, 128*1024L },
		{   2,  32, 128,   F, 2048,   1, 128*1024L },
		{   2,  32, 128,   F, 2048,   1, 256*1024L },
		{   2,   F,   F,   F, 4096,   1, N }
		/* max. compression */
	};

	if (level < 1 || level > 10)
		return GBA_E_INVALID_ARGUMENT;
	sc = &swd_config[level - 1];

	memset(c, 0, sizeof(*c));
	c->ip = c->in = in;
	c->in_end = in + in_len;
	c->out = out;
	cb = NULL;
	c->result = result ? result : (u32 *) result_buffer;
	memset(c->result, 0, 16*sizeof(*c->result));
	c->result[0] = c->result[2] = c->result[4] = UINT_MAX;
	result = NULL;
	memset(&c->conf, 0xff, sizeof(c->conf));
	if (conf)
		memcpy(&c->conf, conf, sizeof(c->conf));
	conf = NULL;
	r = bbConfig(c, 0, 8);
	if (r == 0)
		r = bbConfig(c, c->conf.bb_endian, c->conf.bb_size);
	if (r != 0)
		return GBA_E_INVALID_ARGUMENT;
	c->bb_op = out;

	ii = c->ip;			 /* point to start of literal run */
	lit = 0;

	swd->f = GBA_MIN(F, c->conf.max_match);
	swd->n = GBA_MIN(N, sc->max_offset);
	if (c->conf.max_offset != UINT_MAX)
		swd->n = GBA_MIN(N, c->conf.max_offset);
	if (in_len >= 256 && in_len < swd->n)
		swd->n = in_len;
	if (swd->f < 8 || swd->n < 256)
		return GBA_E_INVALID_ARGUMENT;
	r = init_match(c,swd,NULL,0,sc->flags);
	if (r != GBA_E_OK)
	{
		return r;
	}
	if (sc->max_chain > 0)
		swd->max_chain = sc->max_chain;
	if (sc->nice_length > 0)
		swd->nice_length = sc->nice_length;
	if (c->conf.max_match < swd->nice_length)
		swd->nice_length = c->conf.max_match;

	c->last_m_off = 1;
	r = find_match(c,swd,0,0);
	if (r != GBA_E_OK)
		return r;
	while (c->look > 0)
	{
		u32 ahead;
		u32 max_ahead;
		int l1, l2;

		c->codesize = c->bb_op - out;

		m_len = c->m_len;
		m_off = c->m_off;

		if (lit == 0)
			ii = c->bp;

		if (m_len < 2 || (m_len == 2 && (m_off > M2_MAX_OFFSET))
			|| m_off > c->conf.max_offset)
		{
			/* a literal */
			lit++;
			swd->max_chain = sc->max_chain;
			r = find_match(c,swd,1,0);
			continue;
		}

	/* a match */
		/* shall we try a lazy match ? */
		ahead = 0;
		if (sc->try_lazy <= 0 || m_len >= sc->max_lazy || m_off == c->last_m_off)
		{
			/* no */
			l1 = 0;
			max_ahead = 0;
		}
		else
		{
			/* yes, try a lazy match */
			l1 = len_of_coded_match(c,m_len,m_off);
			max_ahead = GBA_MIN(sc->try_lazy, m_len - 1);
		}

		while (ahead < max_ahead && c->look > m_len)
		{
			if (m_len >= sc->good_length)
				swd->max_chain = sc->max_chain >> 2;
			else
				swd->max_chain = sc->max_chain;
			r = find_match(c,swd,1,0);
			ahead++;

			if (c->m_len < 2)
				continue;
			l2 = len_of_coded_match(c,c->m_len,c->m_off);
			if (l2 < 0)
				continue;

			if (l1 + (int)(ahead + c->m_len - m_len) * 5 > l2 + (int)(ahead) * 9)
			{
				c->lazy++;
				assert_match(swd,c->m_len,c->m_off);

				lit += ahead;

				goto lazy_match_done;
			}
		}

		/* 1 - code run */
		code_run(c,ii,lit);
		lit = 0;

		/* 2 - code match */
		code_match(c,m_len,m_off);
		swd->max_chain = sc->max_chain;
		r = find_match(c,swd,m_len,1+ahead);

lazy_match_done: ;
	}

	/* store final run */
	code_run(c,ii,lit);

	/* EOF */
	bbPutBit(c, 0);
	code_prefix_ss11(c, UINT32_C(0x1000000));
	bbPutByte(c, 0xff);
	bbFlushBits(c, 0);

	c->codesize = c->bb_op - out;
	*out_len = c->bb_op - out;

	swd_exit(swd);
	return GBA_E_OK;
}

